%% this script runs stereo matching from two rectified stereo images
clear;
close all;
clc;
addpath('./plot');

%% specify dataset and parameters
dataset = 'playroom'; img_format = 'png';
img0 = strcat(dataset, '0.', img_format);
img1 = strcat(dataset, '1.', img_format);

load(strcat(dataset, '_calib'));

half_win_size = 50;
win_size = 2* half_win_size + 1;

%% calculate the perspective matrix from the camera parameters
M0 = projectionMatrix(f0, c0, eye(3), zeros(3,1));
M1 = projectionMatrix(f1, c1, eye(3), [-baseline;0;0]); % DELETE

%% load and display
I0 = imread(img0); I0 = rgb2gray(I0);
I1 = imread(img1); I1 = rgb2gray(I1);
I0_pad = padarray(I0, [half_win_size, half_win_size], 'replicate');
I1_pad = padarray(I1, [half_win_size, half_win_size], 'replicate');

f_stereo = figure('Name', 'stereo pair', 'NumberTitle', 'off');
ax_im0 = subplot(121); h0 = imshow(I0); axis on;
ax_im1 = subplot(122); h1 = imshow(I1); axis on;

methods = {'ZNCC', 'SAD', 'SSD'};
f_err = figure('Name', 'epipolar search error', 'NumberTitle', 'off');
ax_err_1 = subplot(311); title(cell2mat(methods(1))); axis on;
ax_err_2 = subplot(312); title(cell2mat(methods(2))); axis on;
ax_err_3 = subplot(313); title(cell2mat(methods(3))); axis on;

aver_range = 5.0;
f_tri = figure('Name', 'triangulation', 'NumberTitle', 'off');
plotCoordinateFrame(eye(3), [0;0;0], 0.5*baseline);
plotCoordinateFrame(eye(3), [baseline;0;0], 0.5*baseline);
rotate3d on; grid on; axis equal; hold on;
xlabel('x'); ylabel('y'); zlabel('z');
view([0, -50, 0]);
ax_tri = gca;
axis(ax_tri, [-aver_range/2, aver_range/2, -aver_range/2, aver_range/2, 0,aver_range]);

%% interactive show epipolar search
disp('starting interactive epipolar search...');
disp('- left click on the left image to see the result of the epipolar search.');
disp('- press "s" in the stereo figure to stop.');
last_pt = [-1;-1];
while 1    
    % check last key pressed
    input_char = get(f_stereo, 'CurrentCharacter');
    if input_char == 's'
        break;
    end
    
    % check last point: only use integer point
    px0 = get(ax_im0, 'CurrentPoint');
    px0 = px0(1, 1:2)'; px0 = round(px0);
    if norm(last_pt-px0)<1 || px0(1)<1 || px0(2)<1 || ...
            px0(1)>img_size(1) || px0(2)>img_size(2)
        pause(0.5);
        continue;
    end
    
    % do epipolar search
    [match_idx, epi1, errors] = searchEpipolar(I0_pad, I1_pad, px0, half_win_size, D_range, methods);
    px1 = epi1(:, match_idx(1));  % take the result of the first measurement
    
    % triangulation
    pt = triangulatePoint(px0, px1, M0, M1); % implement
    
    % calculate reprojection error
    rep_px0 = project3d(pt, M0);
    err_px0 = norm(rep_px0 - px0);
    rep_px1 = project3d(pt, M1);
    err_px1 = norm(rep_px1 - px1);
    
    % redraw
    cla(ax_im0);
    imshow(I0, 'Parent', ax_im0); hold(ax_im0, 'on');
    rectangle('Position', [px0(1)-half_win_size, px0(2)-half_win_size, win_size, win_size], ...
        'Parent', ax_im0, 'LineWidth', 1.0, 'EdgeColor', 'g');
    cla(ax_im1);
    imshow(I1, 'Parent', ax_im1); hold(ax_im1, 'on');
    plot(ax_im1, epi1(1,:), epi1(2,:));
    rectangle('Position', [px1(1)-half_win_size, px1(2)-half_win_size, win_size, win_size], ...
        'Parent', ax_im1, 'LineWidth', 1.0, 'EdgeColor', 'g');
    
    cla(ax_err_1);
    plot(ax_err_1, match_idx(1), errors(1, match_idx(1)), 'sg'); hold(ax_err_1, 'on');
    plot(ax_err_1, errors(1, :)); title(ax_err_1, cell2mat(methods(1)));
    cla(ax_err_2);
    plot(ax_err_2, match_idx(2), errors(2, match_idx(2)), 'sr'); hold(ax_err_2, 'on');
    plot(ax_err_2, errors(2, :)); title(ax_err_2, cell2mat(methods(2)));
    cla(ax_err_3);
    plot(ax_err_3, match_idx(3), errors(3, match_idx(3)), 'sc'); hold(ax_err_3, 'on');
    plot(ax_err_3, errors(3, :)); title(ax_err_3, cell2mat(methods(3)));
    
    plot3(ax_tri, pt(1), pt(2), pt(3), 'Marker', 'o', 'MarkerFaceColor', 'g');
    
    last_pt = px0;
    disp(strcat('last point clicked: ',...
        num2str(last_pt(1)), ',', num2str(last_pt(2))));
    disp(strcat('triangulated point: ', mat2str(pt)));
    disp(strcat('reprojection error in the left image: ', num2str(err_px0)));
    disp(strcat('reprojection error in the right image: ', num2str(err_px1)));
    pause(0.5);
end

